@@ -17,7 +17,6 @@
 package mobac.program.tilestore.berkeleydb;
 
 import java.awt.Point;
-import java.awt.Rectangle;
 import java.awt.image.BufferedImage;
 import java.awt.image.IndexColorModel;
 import java.awt.image.WritableRaster;
@@ -34,6 +33,8 @@ import java.util.Map;
 import java.util.TreeMap;
 
 import mobac.exceptions.TileStoreException;
+import mobac.program.interfaces.MapSource;
+import mobac.program.model.Settings;
 import mobac.program.tilestore.TileStore;
 import mobac.program.tilestore.TileStoreEntry;
 import mobac.program.tilestore.TileStoreInfo;
@@ -42,8 +43,7 @@ import mobac.utilities.GUIExceptionHandler;
 import mobac.utilities.Utilities;
 import mobac.utilities.file.DeleteFileFilter;
 import mobac.utilities.file.DirInfoFileFilter;
-
-import org.openstreetmap.gui.jmapviewer.interfaces.MapSource;
+import mobac.utilities.file.DirectoryFileFilter;
 
 import com.sleepycat.je.DatabaseException;
 import com.sleepycat.je.Environment;
@@ -102,6 +102,7 @@ public class BerkeleyDbTileStore extends TileStore {
 
 		// for (Renamer r : mutations.getRenamers())
 		// log.debug(r.toString());
+		Runtime.getRuntime().addShutdownHook(new ShutdownThread(true));
 	}
 
 	protected void acquireTileStoreLock() throws TileStoreException {
@@ -112,8 +113,8 @@ public class BerkeleyDbTileStore extends TileStore {
 				try {
 					Utilities.mkDirs(tileStoreDir);
 				} catch (IOException e) {
-					throw new TileStoreException("Unable to create tile store directory: \""
-							+ tileStoreDir.getPath() + "\"");
+					throw new TileStoreException("Unable to create tile store directory: \"" + tileStoreDir.getPath()
+							+ "\"");
 				}
 			FileChannel channel = new RandomAccessFile(file, "rw").getChannel();
 
@@ -129,7 +130,7 @@ public class BerkeleyDbTileStore extends TileStore {
 
 			// // Release the lock
 			// lock.release();
-			//	    
+			//
 			// // Close the file
 			// channel.close();
 		} catch (Exception e) {
@@ -139,110 +140,142 @@ public class BerkeleyDbTileStore extends TileStore {
 	}
 
 	@Override
-	public TileStoreEntry createNewEntry(int x, int y, int zoom, byte[] data,
-			long timeLastModified, long timeExpires, String eTag) {
+	public TileStoreEntry createNewEntry(int x, int y, int zoom, byte[] data, long timeLastModified, long timeExpires,
+			String eTag) {
 		return new TileDbEntry(x, y, zoom, data, timeLastModified, timeExpires, eTag);
 	}
 
+	@Override
+	public TileStoreEntry createNewEmptyEntry(int x, int y, int zoom) {
+		long time = System.currentTimeMillis();
+		long timeExpires = time + Settings.getInstance().tileDefaultExpirationTime;
+		// We set the tile data to an empty array because we can not store null
+		return new TileDbEntry(x, y, zoom, new byte[] {}, time, timeExpires, "");
+	}
+
 	private TileDatabase getTileDatabase(MapSource mapSource) throws DatabaseException {
 		TileDatabase db;
 		if (tileDbMap == null)
 			// Tile store has been closed already
 			return null;
+		String storeName = mapSource.getName();
+		if (storeName == null)
+			return null;
 		synchronized (tileDbMap) {
-			db = tileDbMap.get(mapSource.getStoreName());
+			db = tileDbMap.get(storeName);
 		}
 		if (db != null)
 			return db;
 		try {
 			synchronized (tileDbMap) {
 				cleanupDatabases();
-				db = tileDbMap.get(mapSource.getStoreName());
+				db = tileDbMap.get(storeName);
 				if (db == null) {
-					db = new TileDatabase(mapSource);
+					db = new TileDatabase(storeName);
 					db.lastAccess = System.currentTimeMillis();
-					tileDbMap.put(mapSource.getStoreName(), db);
+					tileDbMap.put(mapSource.getName(), db);
 				}
 				return db;
 			}
 		} catch (Exception e) {
-			log.error("Error creating tile store db \"" + mapSource.getStoreName() + "\"", e);
+			log.error("Error creating tile store db \"" + mapSource.getName() + "\"", e);
+			throw new TileStoreException(e);
+		}
+	}
+
+	private TileDatabase getTileDatabase(String storeName) throws DatabaseException {
+		TileDatabase db;
+		if (tileDbMap == null)
+			// Tile store has been closed already
+			return null;
+		if (storeName == null)
+			return null;
+		synchronized (tileDbMap) {
+			db = tileDbMap.get(storeName);
+		}
+		if (db != null)
+			return db;
+		try {
+			synchronized (tileDbMap) {
+				cleanupDatabases();
+				db = tileDbMap.get(storeName);
+				if (db == null) {
+					db = new TileDatabase(storeName);
+					db.lastAccess = System.currentTimeMillis();
+					tileDbMap.put(storeName, db);
+				}
+				return db;
+			}
+		} catch (Exception e) {
+			log.error("Error creating tile store db \"" + storeName + "\"", e);
 			throw new TileStoreException(e);
 		}
 	}
 
 	@Override
-	public TileStoreInfo getStoreInfo(MapSource mapSource) throws InterruptedException {
-		int tileCount = getNrOfTiles(mapSource);
-		long storeSize = getStoreSize(mapSource);
+	public TileStoreInfo getStoreInfo(String storeName) throws InterruptedException {
+		int tileCount = getNrOfTiles(storeName);
+		long storeSize = getStoreSize(storeName);
 		return new TileStoreInfo(storeSize, tileCount);
 	}
 
 	@Override
-	public void putTileData(byte[] tileData, int x, int y, int zoom, MapSource mapSource)
-			throws IOException {
+	public void putTileData(byte[] tileData, int x, int y, int zoom, MapSource mapSource) throws IOException {
 		this.putTileData(tileData, x, y, zoom, mapSource, -1, -1, null);
 	}
 
 	@Override
-	public void putTileData(byte[] tileData, int x, int y, int zoom, MapSource mapSource,
-			long timeLastModified, long timeExpires, String eTag) throws IOException {
-		if (!mapSource.allowFileStore())
-			return;
-		TileDbEntry tile = new TileDbEntry(x, y, zoom, tileData, timeLastModified, timeExpires,
-				eTag);
+	public void putTileData(byte[] tileData, int x, int y, int zoom, MapSource mapSource, long timeLastModified,
+			long timeExpires, String eTag) throws IOException {
+		TileDbEntry tile = new TileDbEntry(x, y, zoom, tileData, timeLastModified, timeExpires, eTag);
 		TileDatabase db = null;
 		try {
 			if (log.isTraceEnabled())
-				log.trace("Saved " + mapSource.getStoreName() + " " + tile);
+				log.trace("Saved " + mapSource.getName() + " " + tile);
 			db = getTileDatabase(mapSource);
 			if (db != null)
 				db.put(tile);
 		} catch (Exception e) {
 			if (db != null)
 				db.close();
-			log.error("Faild to write tile to tile store \"" + mapSource.getStoreName() + "\"", e);
+			log.error("Faild to write tile to tile store \"" + mapSource.getName() + "\"", e);
 		}
 	}
 
 	@Override
 	public void putTile(TileStoreEntry tile, MapSource mapSource) {
-		if (!mapSource.allowFileStore())
-			return;
 		TileDatabase db = null;
 		try {
 			if (log.isTraceEnabled())
-				log.trace("Saved " + mapSource.getStoreName() + " " + tile);
+				log.trace("Saved " + mapSource.getName() + " " + tile);
 			db = getTileDatabase(mapSource);
 			db.put((TileDbEntry) tile);
 		} catch (Exception e) {
 			if (db != null)
 				db.close();
-			log.error("Faild to write tile to tile store \"" + mapSource.getStoreName() + "\"", e);
+			log.error("Faild to write tile to tile store \"" + mapSource.getName() + "\"", e);
 		}
 	}
 
 	@Override
 	public TileStoreEntry getTile(int x, int y, int zoom, MapSource mapSource) {
-		if (!mapSource.allowFileStore())
-			return null;
 		TileDatabase db = null;
 		try {
 			db = getTileDatabase(mapSource);
+			if (db == null)
+				return null;
 			TileStoreEntry tile = db.get(new TileDbKey(x, y, zoom));
 			if (log.isTraceEnabled()) {
 				if (tile == null)
-					log.trace("Tile store cache miss: (x,y,z)" + x + "/" + y + "/" + zoom + " "
-							+ mapSource.getStoreName());
+					log.trace("Tile store cache miss: (x,y,z)" + x + "/" + y + "/" + zoom + " " + mapSource.getName());
 				else
-					log.trace("Loaded " + mapSource.getStoreName() + " " + tile);
+					log.trace("Loaded " + mapSource.getName() + " " + tile);
 			}
 			return tile;
 		} catch (Exception e) {
 			if (db != null)
 				db.close();
-			log.error("failed to retrieve tile from tile store \"" + mapSource.getStoreName()
-					+ "\"", e);
+			log.error("failed to retrieve tile from tile store \"" + mapSource.getName() + "\"", e);
 			return null;
 		}
 	}
@@ -257,47 +290,44 @@ public class BerkeleyDbTileStore extends TileStore {
 	}
 
 	public void prepareTileStore(MapSource mapSource) {
-		if (!mapSource.allowFileStore())
-			return;
 		try {
 			getTileDatabase(mapSource);
 		} catch (DatabaseException e) {
 		}
 	}
 
-	public void clearStore(MapSource mapSource) {
-		File databaseDir = getStoreDir(mapSource);
+	public void clearStore(String storeName) {
+		File databaseDir = getStoreDir(storeName);
 
 		TileDatabase db;
 		synchronized (tileDbMap) {
-			db = tileDbMap.get(mapSource.getStoreName());
+			db = tileDbMap.get(storeName);
 			if (db != null)
 				db.close(false);
 			if (databaseDir.exists()) {
 				DeleteFileFilter dff = new DeleteFileFilter();
 				databaseDir.listFiles(dff);
 				databaseDir.delete();
-				log.debug("Tilestore " + mapSource.getStoreName() + " cleared: " + dff);
+				log.debug("Tilestore " + storeName + " cleared: " + dff);
 			}
-			tileDbMap.remove(mapSource.getStoreName());
+			tileDbMap.remove(storeName);
 		}
 	}
 
 	/**
-	 * This method returns the amount of tiles in the store of tiles which is
-	 * specified by the {@link MapSource} object.
+	 * This method returns the amount of tiles in the store of tiles which is specified by the {@link MapSource} object.
 	 * 
-	 * @param mapSource
+	 * @param mapSourceName
 	 *            the store to calculate number of tiles in
 	 * @return the amount of tiles in the specified store.
 	 * @throws InterruptedException
 	 */
-	public int getNrOfTiles(MapSource mapSource) throws InterruptedException {
+	public int getNrOfTiles(String mapSourceName) throws InterruptedException {
 		try {
-			File storeDir = getStoreDir(mapSource);
+			File storeDir = getStoreDir(mapSourceName);
 			if (!storeDir.isDirectory())
 				return 0;
-			TileDatabase db = getTileDatabase(mapSource);
+			TileDatabase db = getTileDatabase(mapSourceName);
 			int tileCount = (int) db.entryCount();
 			db.close();
 			return tileCount;
@@ -307,8 +337,8 @@ public class BerkeleyDbTileStore extends TileStore {
 		}
 	}
 
-	public long getStoreSize(MapSource mapSource) throws InterruptedException {
-		File tileStore = getStoreDir(mapSource);
+	public long getStoreSize(String storeName) throws InterruptedException {
+		File tileStore = getStoreDir(storeName);
 		if (tileStore.exists()) {
 			DirInfoFileFilter diff = new DirInfoFileFilter();
 			try {
@@ -322,8 +352,8 @@ public class BerkeleyDbTileStore extends TileStore {
 		}
 	}
 
-	public BufferedImage getCacheCoverage(MapSource mapSource, int zoom, Point tileNumMin,
-			Point tileNumMax) {
+	public BufferedImage getCacheCoverage(MapSource mapSource, int zoom, Point tileNumMin, Point tileNumMax)
+			throws InterruptedException {
 		TileDatabase db;
 		try {
 			db = getTileDatabase(mapSource);
@@ -352,30 +382,8 @@ public class BerkeleyDbTileStore extends TileStore {
 		}
 	}
 
-	public void closeAll(final boolean shutdown) {
-		Thread t = new DelayedInterruptThread("DBShutdown") {
-
-			@Override
-			public void run() {
-				log.debug("Closing all tile databases...");
-				synchronized (tileDbMap) {
-					for (TileDatabase db : tileDbMap.values()) {
-						db.close(false);
-					}
-					tileDbMap.clear();
-					if (shutdown) {
-						tileDbMap = null;
-						try {
-							tileStoreLock.release();
-						} catch (IOException e) {
-							log.error("", e);
-						}
-					}
-				}
-				log.debug("All tile databases has been closed");
-			}
-
-		};
+	public void closeAll() {
+		Thread t = new ShutdownThread(false);
 		t.start();
 		try {
 			t.join();
@@ -385,8 +393,7 @@ public class BerkeleyDbTileStore extends TileStore {
 	}
 
 	/**
-	 * Returns <code>true</code> if the tile store directory of the specified
-	 * {@link MapSource} exists.
+	 * Returns <code>true</code> if the tile store directory of the specified {@link MapSource} exists.
 	 * 
 	 * @param mapSource
 	 * @return
@@ -397,17 +404,72 @@ public class BerkeleyDbTileStore extends TileStore {
 	}
 
 	/**
+	 * Returns the directory used for storing the tile database of the {@link MapSource} specified by
+	 * <code>mapSource</code>
+	 * 
 	 * @param mapSource
-	 * @return directory used for storing the tile database belonging to
-	 *         <code>mapSource</code>
+	 * @return
 	 */
 	protected File getStoreDir(MapSource mapSource) {
-		return new File(tileStoreDir, "db-" + mapSource.getStoreName());
+		return getStoreDir(mapSource.getName());
+	}
+
+	/**
+	 * @param mapSourceName
+	 * @return directory used for storing the tile database belonging to <code>mapSource</code>
+	 */
+	protected File getStoreDir(String mapSourceName) {
+		return new File(tileStoreDir, "db-" + mapSourceName);
+	}
+
+	public String[] getAllStoreNames() {
+		File[] dirs = tileStoreDir.listFiles(new DirectoryFileFilter());
+		ArrayList<String> storeNames = new ArrayList<String>(dirs.length);
+		for (File d : dirs) {
+			String name = d.getName();
+			if (name.startsWith("db-")) {
+				name = name.substring(3);
+				storeNames.add(name);
+			}
+		}
+		String[] result = new String[storeNames.size()];
+		storeNames.toArray(result);
+		return result;
+	}
+
+	private class ShutdownThread extends DelayedInterruptThread {
+
+		private final boolean shutdown;
+
+		public ShutdownThread(boolean shutdown) {
+			super("DBShutdown");
+			this.shutdown = shutdown;
+		}
+
+		@Override
+		public void run() {
+			log.debug("Closing all tile databases...");
+			synchronized (tileDbMap) {
+				for (TileDatabase db : tileDbMap.values()) {
+					db.close(false);
+				}
+				tileDbMap.clear();
+				if (shutdown) {
+					tileDbMap = null;
+					try {
+						tileStoreLock.release();
+					} catch (IOException e) {
+						log.error("", e);
+					}
+				}
+			}
+			log.debug("All tile databases has been closed");
+		}
 	}
 
 	protected class TileDatabase {
 
-		final MapSource mapSource;
+		final String mapSourceName;
 		final Environment env;
 		final EntityStore store;
 		final PrimaryIndex<TileDbKey, TileDbEntry> tileIndex;
@@ -415,16 +477,15 @@ public class BerkeleyDbTileStore extends TileStore {
 
 		long lastAccess;
 
-		public TileDatabase(MapSource mapSource) throws IOException, EnvironmentLockedException,
-				DatabaseException {
-			log.debug("Opening tile store db: \"" + mapSource.getStoreName() + "\"");
+		public TileDatabase(String mapSourceName) throws IOException, EnvironmentLockedException, DatabaseException {
+			log.debug("Opening tile store db: \"" + mapSourceName + "\"");
 			DelayedInterruptThread t = (DelayedInterruptThread) Thread.currentThread();
 			try {
 				t.pauseInterrupt();
-				this.mapSource = mapSource;
+				this.mapSourceName = mapSourceName;
 				lastAccess = System.currentTimeMillis();
 
-				File storeDir = getStoreDir(mapSource);
+				File storeDir = getStoreDir(mapSourceName);
 				Utilities.mkDirs(storeDir);
 
 				env = new Environment(storeDir, envConfig);
@@ -441,7 +502,7 @@ public class BerkeleyDbTileStore extends TileStore {
 					close();
 				t.resumeInterrupt();
 			}
-			log.debug("Opened tile store db: \"" + mapSource.getStoreName() + "\"");
+			log.debug("Opened tile store db: \"" + mapSourceName + "\"");
 		}
 
 		public boolean isClosed() {
@@ -472,8 +533,9 @@ public class BerkeleyDbTileStore extends TileStore {
 			return tileIndex.get(key);
 		}
 
-		public BufferedImage getCacheCoverage(int zoom, Point tileNumMin, Point tileNumMax)
-				throws DatabaseException {
+		public BufferedImage getCacheCoverage(int zoom, Point tileNumMin, Point tileNumMax) throws DatabaseException,
+				InterruptedException {
+			log.debug("Loading cache coverage for region " + tileNumMin + " " + tileNumMax + " of zoom level " + zoom);
 			DelayedInterruptThread t = (DelayedInterruptThread) Thread.currentThread();
 			int width = tileNumMax.x - tileNumMin.x + 1;
 			int height = tileNumMax.y - tileNumMin.y + 1;
@@ -484,8 +546,7 @@ public class BerkeleyDbTileStore extends TileStore {
 			IndexColorModel colorModel = new IndexColorModel(2, 2, colors, 0, true);
 			BufferedImage image = null;
 			try {
-				image = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_INDEXED,
-						colorModel);
+				image = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_INDEXED, colorModel);
 			} catch (Throwable e) {
 				log.error("Failed to create coverage image: " + e.toString());
 				image = null;
@@ -494,22 +555,27 @@ public class BerkeleyDbTileStore extends TileStore {
 			}
 			WritableRaster raster = image.getRaster();
 
-			TileDbKey fromKey = new TileDbKey(tileNumMin.x, tileNumMin.y, zoom);
-			TileDbKey toKey = new TileDbKey(tileNumMax.x, tileNumMax.y, zoom);
-			EntityCursor<TileDbKey> cursor = tileIndex.keys(fromKey, true, toKey, true);
-			try {
-				Rectangle r = new Rectangle(tileNumMin.x, tileNumMin.y, width, height);
-				TileDbKey key = cursor.next();
-				while (key != null) {
-					if (r.contains(key.x, key.y)) {
-						raster.setSample(key.x - tileNumMin.x, key.y - tileNumMin.y, 0, 1);
+			// We are loading the coverage of the selected area column by column which is much faster than loading the
+			// whole region at once
+			for (int x = tileNumMin.x; x <= tileNumMax.x; x++) {
+				TileDbKey fromKey = new TileDbKey(x, tileNumMin.y, zoom);
+				TileDbKey toKey = new TileDbKey(x, tileNumMax.y, zoom);
+				EntityCursor<TileDbKey> cursor = tileIndex.keys(fromKey, true, toKey, true);
+				try {
+					TileDbKey key = cursor.next();
+					while (key != null) {
+						int pixelx = key.x - tileNumMin.x;
+						int pixely = key.y - tileNumMin.y;
+						raster.setSample(pixelx, pixely, 0, 1);
+						key = cursor.next();
+						if (t.isInterrupted()) {
+							log.debug("Cache coverage loading aborted");
+							throw new InterruptedException();
+						}
 					}
-					key = cursor.next();
-					if (t.interruptedWhilePaused())
-						return null;
+				} finally {
+					cursor.close();
 				}
-			} finally {
-				cursor.close();
 			}
 			return image;
 		}
@@ -532,16 +598,16 @@ public class BerkeleyDbTileStore extends TileStore {
 				return;
 			if (removeFromMap) {
 				synchronized (tileDbMap) {
-					TileDatabase db2 = tileDbMap.get(mapSource.getStoreName());
+					TileDatabase db2 = tileDbMap.get(mapSourceName);
 					if (db2 == this)
-						tileDbMap.remove(mapSource.getStoreName());
+						tileDbMap.remove(mapSourceName);
 				}
 			}
 			DelayedInterruptThread t = (DelayedInterruptThread) Thread.currentThread();
 			try {
 				t.pauseInterrupt();
 				try {
-					log.debug("Closing tile store db \"" + mapSource.getStoreName() + "\"");
+					log.debug("Closing tile store db \"" + mapSourceName + "\"");
 					if (store != null)
 						store.close();
 				} catch (Exception e) {
